#ifndef __DHCPCLIENT__DHCP_H
#define __DHCPCLIENT__DHCP_H

#include "support.h"

#define DHCP_DISCOVER   1
#define DHCP_OFFER      2
#define DHCP_REQUEST    3
#define DHCP_DECLINE    4
#define DHCP_ACK        5
#define DHCP_NACK       6
#define DHCP_RELEASE    7
#define DHCP_INFORM     8


#define DHCP_REQUEST_SELECT     111
#define DHCP_REQUEST_RESELECT   122
#define DHCP_REQUEST_RENEW      133
#define DHCP_REQUEST_REBIND     144
#define DHCP_RENEW              201

#define DHCP_INIT               1001
#define DHCP_ERROR              1002
#define DHCP_DEBUG              1003

#define DHCP_SERVER_PORT        67
#define DHCP_CLIENT_PORT        68

#define DHCP_C_REQ_BCAST_REPLY_ALWAYS   1
#define DHCP_C_REQ_BCAST_REPLY_NEVER    2
#define DHCP_C_SEND_BCAST_PACKETS       4
#define DHCP_C_CHECK_IP_USE             8
#define DHCP_C_NOCOLLECT_OFFERS         16

class   DHCPMsg {
public:
        unsigned char   op;// 1 = BOOTREQUEST 2 = BOOTREPLY
        unsigned char   htype; //1
        unsigned char   hlen;//6
        unsigned char   hops;//0
        unsigned long   xid; //Trans ID set by client
        unsigned short  secs; //set 0
        unsigned short  flags; //1 if broadcast is required
        unsigned char   ciaddr[4]; //current client address (0) otherwise
        unsigned char   yiaddr[4]; //my client ip upon boot or 0
        unsigned char   siaddr[4]; //returned by server to indicate failure
        unsigned char   giaddr[4]; //relay agent ip (whocares)
        unsigned char   chaddr[16]; //client hardware address
        unsigned char   sname[64]; //server hostname
        unsigned char   file[128]; //boot file
        unsigned char   Cookie[4]; //client cookie
        unsigned char   options[308]; //options 255 for end
};

class   DHCPLeaseData {
public:
        unsigned char           StartData;
        unsigned char           ServerIP[4];
        unsigned char           LocalIP[4];
        unsigned char           Routers[4][4];
        unsigned char           DomainNameServers[4][4];
        unsigned char           Netmask[4];
        unsigned char           HostName[64];
        unsigned char           DomainName[64];
        unsigned char           Broadcast[4];
        unsigned long           RenewTime;

        void    Reset();
};

class   DHCPSessionData {
public:
        unsigned char           HardwareAddress[16];
        unsigned char           ClientID[16];
        int                     CIDLength;
        int                     HALength;
        int                     HAType;
        int                     ConformParms;
        unsigned char           PrimeIP[4];     //primary local ip; ie, the interfaces IP
        char                    Interface[64];  //IP or interface name
        unsigned short          ClientPort;
        unsigned char           BroadcastAddress[4];
        unsigned char           NullcastAddress[4];
        unsigned char           Options[128];
        unsigned char           RequestOptions[256];
        char                    AuthName[64];
        char                    ClassID[64];
        int                     PrefTime;
        unsigned long           DiscoverTimeout;
        unsigned long           RequestTimeout;

        unsigned char           Servers[4][4];
        unsigned char           Offered[4][4];
        int                     nOffers;

        void    Reset();
};

class   DHCP : public DHCPLeaseData, public DHCPSessionData {
protected:
        class   DHCPSocket {
        public:
                int             s;

                virtual int     Send(void *,
                                     int,
                                     unsigned short,
                                     const unsigned char *) = 0;

                virtual int     Recv(void *,
                                     int,
                                     unsigned short &,
                                     unsigned char *) = 0;

                virtual int     CheckDup(const unsigned char *,const unsigned char *) = 0;

                virtual void    Unarp() = 0;

                virtual         ~DHCPSocket();
        };

        class   DHCPUdpSocket : public DHCPSocket {
        public:
                int             Send(void *,
                                     int,
                                     unsigned short,
                                     const unsigned char *);

                int             Recv(void *,
                                     int,
                                     unsigned short &,
                                     unsigned char *);

                int             CheckDup(const unsigned char *,const unsigned char *);

                void            Unarp();

                DHCPUdpSocket(unsigned short,unsigned char *);
        };

#ifndef __WIN32__
        class   DHCPRawSocket : public DHCPSocket {
        private:
                struct PacketInfo {
                        ETHERHdr *eh;
                        IPHdr *ih;
                        UDPHdr *uh;
                        unsigned char *data;
                };
        public:
                char *          Interface;
                unsigned char * HWAddress;
                int             HALength;
                unsigned short  Port;

                unsigned int    ConformParms;

                unsigned char   *IP;
                unsigned char   *Netmask;
                unsigned char   *Gateway;

                const unsigned char * GetTarget(const unsigned char *);

                int             TestPacket(unsigned char *,int,PacketInfo &);

                int             Send(void *,
                                     int,
                                     unsigned short,
                                     const unsigned char *);

                int             Recv(void *,
                                     int,
                                     unsigned short &,
                                     unsigned char *);

                int             CheckDup(const unsigned char *,const unsigned char *);

                void            Unarp();

                DHCPRawSocket(char *,
                              unsigned char *,
                              int,
                              unsigned short,
                              unsigned int,
                              unsigned char *,
                              unsigned char *,
                              unsigned char *);
        };
#endif

        DHCPSocket *            Socket;
        DHCPMsg                 Packet;
        static IDF              RandomID;

        unsigned char * GetOption(int);
        int             WaitResponse(unsigned long,unsigned long);
        unsigned long   PrepRequest();
        void            ConfigureInterface(char *,int,int,const unsigned char *,int);
        void            AddOption(unsigned char,int,const unsigned char *);
        void            AddOption(unsigned char,int,const char *);
        void            AddOption(unsigned char,int,long,...);
        void            AddOption(unsigned char,int,unsigned long,...);
        void            AddOption(unsigned char,int,int,...);
        unsigned long   GetID();
        int             Send(unsigned char *);
        int             Request(const unsigned char *,
                                const unsigned char *,
                                const unsigned char *,
                                int,
                                int);
        void            StoreBasic(unsigned char *,unsigned char *);
        void            StoreOptions();
        void            StoreTime();
        void            StoreList(unsigned char *,unsigned char *,int,int,int);
        int             OpenChannel();
        void            CloseChannel();
        void            Decline(const unsigned char *);
        void            AddOptions(unsigned char *);

        virtual void    Handler(int,char *,char *);

public:

        void    SetConformity(int);
        int     Rebind();
        int     IsSelectee(int,const IPName &);
        int     Select(int);
        int     Select(int,const IPName &);
        int     Select(const IPName &,const IPName &);
        int     Reselect(const IPName &);
        int     Inform();
        int     Discover(const IPName &);
        void    Release();
        int     Renew();
        void    SetPrimeIP(const IPName &);
        void    SetAuthName(char *);
        void    SetClassID(char *);
        unsigned char * ConvertIP(char *);
        int     SetPrefTime(int);
        void    SetOptions(unsigned char *);
        void    AddRequestOptions(unsigned char *);
        void    SetTimeouts(unsigned long,unsigned long);

        DHCP();
        void Init(char *,int,int,const HWName &,int);
        virtual ~DHCP();
};

#endif
