#include "includes.h"
#include "support.h"
#include "packet.h"
#include "dhcp.h"

/* Renew hiarchy
        Renew time elasped in 8ths.

        0/8     Just bound.
        1/8     Sleeping.
        2/8     Sleeping.
        3/8     Sleeping.
        4/8     Sleeping.
        5/8     Begin renewing.
        6/8     Begin rebinding.
        7/8     Assume lease expired.
        8/8     Lease expired.
*/

/* Selecting hiarchy

INITIAL STATES

 +---------------------------------------------------+  +--+
 |                                                   |  |  |
 |     ----------      ------------      ----------  |  |  --------
 +-->--|RESELECT|      |REDISCOVER|      |DISCOVER|  ^  ^  |INFORM|
 |     ----------      ------------      ----------  |  |  --------
 |     |        +-FAIL>+ |        +-FAIL>+   |    +--+  +--+
 |     V                 V        |          V    |   FAIL 
 |  SUCCESS           SUCCESS     |       SUCCESS |  SLEEP 1 MIN
 |     |                 |        ^          |    ^
 |     |               ------------------   ----------------
 |     |               |REDISCOVERSELECT|   |DISCOVERSELECT|
 |     |               ------------------   ----------------
 |     |               V                    V
 |     |            SUCCESS              SUCCESS
 |     |               |                    |
 |     +---------------+--------------------+
 |     |               |                    |
 |     V            SUCCESS              SUCCESS
 |     |               ^                    ^
 |     -------      ----------         -----------         ---------
 |     |BOUND|-->---|RENEWING|-->FAIL--|REBINDING|-->FAIL--|RELEASE|
 |     -------      ----------         -----------         ---------
 |     |SLEEP|                                                 |
 |     -------                                                 |
 |                                                             |
 --------------------------------------------------------------+
*/

int     _argc;
char ** _argv;

char *  MainPath = "/etc/dhcpxd";
char *  ShPath = "/etc/dhcpxd/sh";
char *  LockFile = "/var/lock/dhcpxd.lock";
char    PIDFile[128] = {"/var/run/dhcpxd.pid"};

typedef void (*sighandler)(int);

sighandler      Signal(int signum,sighandler handler)
{
        struct sigaction action;
        sighandler ohandler;

        sigaction(signum,0,&action);
        ohandler = action.sa_handler;
        action.sa_handler = handler;
        action.sa_flags &= ~SA_ONESHOT;
        sigaction(signum,&action,0);

        return ohandler;
}

class   DHCPClient : public DHCP {
public:
        typedef unsigned long (DHCPClient::* DHCPState) (unsigned long);

        struct StartupData {
                StartupData *   next;

                //DHCP:: specific
                char            Interface[64];
                int             Flags;
                char            ClientID[32];
                int             Append;
                char            HostName[64];
                char            ClassID[64];
                unsigned char   Options[128];
                unsigned char   RequestOptions[256];
                unsigned long   Time;
                unsigned long   TimeoutValue;

                //DHCPClient:: specific
                char            TargetInterface[64];
                char            ScriptExtension[16];
                DHCPState       State;
                char            Parameters[128];

                int             TestFlag;
                int             WaitFlag;
                int             IgnoreFlag;
                int             ReleaseFlag;
                int             DeleteFlag;
                int             NackFlag;
        };

        static int              LockX();
        static void             UnlockX(int);
        static int              RenewSession(char *);
        static int              KillSession(char *);
        static int              KillCount;
        static int              AlarmCount;
        static int              TriggerCount;
        static int              WaitCount;
        static DHCPClient *     FirstSession;
        static DHCPClient *     LastSession;
        static int              EventManager();
        static int              ProcessCmdLine(char * (*)(),StartupData *);
        static int              Main(char * (*)());
        static void             KillHandler(int);
        static void             AlarmHandler(int);
        static void             TriggerHandler(int);
        static void             CreatePIDFile(int);
        static void             CFork();
        static int              cmd_alias(char *);

        DHCPClient *    prev;
        DHCPClient *    next;

        unsigned long   StateTimeout;           //For use by bound states
        unsigned long   TickStart;              //Set by StateTransition
        unsigned long   Timeout;                //Used for intra behavior
        unsigned long   Retry;                  //Count down retry counter
        char            PrevIP[20];             //Holder for previous ip

        char            TargetInterface[64];
        char            ScriptExtension[16];
        DHCPState       State;
        DHCPState       TransState;
        DHCPState       AfterSleepState;
        char            Parameters[128];
        const char *    LastAction;

        int             WaitFlag;               //Used to indicate to wait
        int             ReleaseFlag;            //Release the IP back to pool
        int             NackFlag;

        unsigned long   StateDiscover(unsigned long);
        unsigned long   StateRediscover(unsigned long);
        unsigned long   StateReselect(unsigned long);
        unsigned long   StateInform(unsigned long);

        unsigned long   StateDiscoverSelect(unsigned long);
        unsigned long   StateRediscoverSelect(unsigned long);

        unsigned long   StateRenew(unsigned long);
        unsigned long   StateRebind(unsigned long);

        unsigned long   StateSleep(unsigned long);
        unsigned long   StateTransition(unsigned long);
                                                //sets tick start
                                                //sets, calls, and returns
                                                //from the new state

        void            ChangeState(DHCPState);
        void            SleepState(DHCPState);
        void            MorphState(DHCPState);

        int             Test();
        void            Close();

        DHCPClient();
        DHCPClient(StartupData &);
        void            Last();
        void            First();
        ~DHCPClient();

        void            WriteCache();
        void            WriteInfo();
        void            WriteParameters();
        void            ExecScript(const char *);

        void            Handler(int,char *,char *);
};


int     DHCPClient::LockX()
{
        int fd = open(LockFile,O_CREAT|O_RDONLY,0400);
        if(fd == -1) printf("%s: Lock failed!\n",_argv[0]);
        flock(fd,LOCK_EX);
        return fd;
}

void    DHCPClient::UnlockX(int fd)
{
        flock(fd,LOCK_UN);
        close(fd);
}

//#define __EXECMODE_SYSTEM
#define __EXECMODE_SHSTR
//#define __EXECMODE_SHADV

void    DHCPClient::ExecScript(const char *file)
{
        char fname[256];
        int lfd = LockX();
        LastAction = file;

#ifdef  __EXECMODE_SYSTEM
        char *fmt = "\"%s/ifcfg.%s\" \"%s\" \"%s\" \"%s\" \"%s\"";
        sprintf(fname,fmt,MainPath,ScriptExtension,file,TargetInterface,Interface,MainPath);
        system(fname);
#endif

#ifdef  __EXECMODE_SHSTR
        char *fmt = "\"%s/ifcfg.%s\" \"%s\" \"%s\" \"%s\" \"%s\"";
        sprintf(fname,fmt,MainPath,ScriptExtension,file,TargetInterface,Interface,MainPath);

        int pid = fork();
        if(pid == 0) {
                execl(ShPath,ShPath,"-c",fname,(char *)0);
                exit(1);
        };
        if(pid > 0) waitpid(pid,0,0);
#endif

#ifdef  __EXECMODE_SHADV
        sprintf(fname,"%s/ifcfg.%s",MainPath,ScriptExtension);

        int pid = fork();
        if(pid == 0) {
                execl(ShPath,ShPath,"--",fname,file,TargetInterface,Interface,MainPath,0);
                exit(1);
        };
        if(pid > 0) waitpid(pid,0,0);
#endif

        UnlockX(lfd);
}

void    DHCPClient::CFork()
{
        int pid;

        alarm(0);
        AlarmCount = 0;
        pid = fork();
        if(pid > 0) {
                while(AlarmCount == 0) pause();
                while(FirstSession) delete FirstSession;
                exit(0);
        };
        if(pid == 0) {
                CreatePIDFile(getpid());

                setsid();
                int fd;

                fd = open("/dev/null",O_CREAT, 0666|O_APPEND|O_NOCTTY|O_SYNC|O_RDONLY);
                dup2(fd,0);
                close(fd);

                char fname[129],sym[129];
                int n;
                sprintf(fname,"%s/out",MainPath);
                fd = open(fname,O_CREAT, 0666|O_APPEND|O_NOCTTY|O_SYNC|O_WRONLY);
                if(fd != -1) {
                        strcpy(sym,fname);
                        n = readlink(fname,sym,128);
                        if(n > 0) sym[n] = 0;
                        printf("Switch to console %s\n",sym);
                        dup2(fd,1);
                        close(fd);
                        printf("Switch to console %s\n",sym);
                };

                fd = open("/dev/tty1",O_CREAT, 0666|O_APPEND|O_NOCTTY|O_SYNC|O_WRONLY);
                dup2(fd,2);
                close(fd);

                kill(getppid(),SIGALRM);
        };
        if(pid < 0) CreatePIDFile(getpid());
}

void    DHCPClient::CreatePIDFile(int pid)
{
        int lfd = LockX();

        FILE *f;

        f = fopen(PIDFile,"w+");
        if(f) {
                fprintf(f,"%d",pid);
                fclose(f);
        };

        UnlockX(lfd);
}

void    DHCPClient::WriteCache()
{
        char fname[128];
        FILE *f;
        sprintf(fname,"%s/%s.cache",MainPath,TargetInterface);
        f = fopen(fname,"w+");
        if(f) {
                fprintf(f,"%s",PrevIP);
                fclose(f);
        };
}

void    DHCPClient::WriteParameters()
{
        char fname[128];
        FILE *f;
        sprintf(fname,"%s/%s.pars",MainPath,TargetInterface);
        f = fopen(fname,"w+");
        if(f) {
                fprintf(f,"%s",Parameters);
                fclose(f);
        };
}

void    DHCPClient::WriteInfo()
{
        unsigned char Network[4];
        int c;
        char fname[128];
        FILE *f;
        sprintf(fname,"%s/%s.info",MainPath,TargetInterface);
        f = fopen(fname,"w+");
        if(f) {
                c = 0;
                while(c < 4) {
                        Network[c] = LocalIP[c] & Netmask[c];
                        c++;
                };

                fprintf(f,"INTERFACE=\"%s\"\n",TargetInterface);
                fprintf(f,"BASEINTERFACE=\"%s\"\n",Interface);
                fprintf(f,"IP=\"%s\"\n",ipname(LocalIP));
                fprintf(f,"SERVER=\"%s\"\n",ipname(ServerIP));
                fprintf(f,"HOSTNAME=\"%s\"\n",HostName);
                fprintf(f,"DOMAIN=\"%s\"\n",DomainName);
                fprintf(f,"NETWORK=\"%s\"\n",ipname(Network));
                fprintf(f,"NETMASK=\"%s\"\n",ipname(Netmask));
                fprintf(f,"BROADCAST=\"%s\"\n",ipname(Broadcast));

                fprintf(f,"GATEWAY=\"");
                c = 0;
                while(c < 4 && memcmp(Routers[c],"\x0\x0\x0\x0",4) != 0) {
                        if(c > 0) fprintf(f," ");
                        fprintf(f,"%s",ipname(Routers[c]));
                        c++;
                };
                fprintf(f,"\"\n");

                fprintf(f,"DNS=\"");
                c = 0;
                while(c < 4 && memcmp(DomainNameServers[c],"\x0\x0\x0\x0",4) != 0) {
                        if(c > 0) fprintf(f," ");
                        fprintf(f,"%s",ipname(DomainNameServers[c]));
                        c++;
                };
                fprintf(f,"\"\n");

                fprintf(f,"PID=\"%d\"\n",getpid());
                fprintf(f,"WAITFLAG=\"%d\"\n",WaitFlag);
                fprintf(f,"WAITCOUNT=\"%d\"\n",WaitCount);

                unsigned char *ptr = Packet.options;
                while(*ptr != 255 && (ptr - Packet.options) < 308) {
                        if(!*ptr) {
                                ptr++;
                                continue;
                        };
                        int opnum = *ptr;
                        ptr++;
                        int len = *ptr;
                        ptr++;
                        if((len % 4) == 0) {
                                unsigned char *data = ptr;
                                fprintf(f,"OPTION_%03u=\"",opnum);
                                while((data - ptr) < len) {
                                        if(data > ptr) fprintf(f," ");
                                        fprintf(f,"%s",ipname(data));
                                        data += 4;
                                };
                                fprintf(f,"\"\n");
                        };
                        ptr += len;
                };

                fclose(f);
        };
}


int     DHCPClient::TriggerCount = 0;

void    DHCPClient::TriggerHandler(int)
{
        TriggerCount++;
}

int     DHCPClient::AlarmCount = 0;

void    DHCPClient::AlarmHandler(int)
{
        AlarmCount++;
}

int     DHCPClient::KillCount = 0;

void    DHCPClient::KillHandler(int)
{
        KillCount++;
}

int     DHCPClient::Test()
{
        if(Discover(0) <= 0) return 1;
        if(Select(0) <= 0) return 2;

        unsigned char ip[4];

        Renew();
        Rebind();
        memcpy(ip,LocalIP,4);
        Inform();
        Release();

        if(Discover(ipname(ip)) <= 0 ||
           Select(0,ipname(ip)) <= 0) {
           return 3;
        };

        if(Reselect(ipname(ip)) <= 0) {
           Release();
           return 4;
        };

        Release();
        return 5;
}

void    DHCPClient::ChangeState(DHCPState state)
{
        TransState = state;
        State = &DHCPClient::StateTransition;
}

void    DHCPClient::SleepState(DHCPState state)
{
        AfterSleepState = state;
        TransState = &DHCPClient::StateSleep;
        State = &DHCPClient::StateTransition;       
}

void    DHCPClient::MorphState(DHCPState state)
{
        State = state;
}




DHCPClient::DHCPClient()
{
#ifdef  __WIN32__
        Init("24.64.221.215",0,0,"00:02:67:09:05:07",0);
        strcpy(TargetInterface,"24.64.221.215");
#else
        Init("eth0",0,0,0,0);
        strcpy(TargetInterface,"eth0");
#endif


        next = 0;
        prev = LastSession;
        if(LastSession == 0) FirstSession = this;
        else LastSession[0].next = this;
        LastSession = this;
}




DHCPClient *    DHCPClient::FirstSession = 0;
DHCPClient *    DHCPClient::LastSession = 0;
int             DHCPClient::WaitCount = 0;

DHCPClient::DHCPClient(StartupData &sd)
{
        LastAction = "";

        strcpy(TargetInterface,sd.TargetInterface);
        strcpy(ScriptExtension,sd.ScriptExtension);
        memcpy(Parameters,sd.Parameters,128);
        strcpy(Interface,sd.Interface);

        WriteParameters();
        ExecScript("init");

        WaitFlag = sd.WaitFlag;
        ReleaseFlag = sd.ReleaseFlag;
        NackFlag = sd.NackFlag;

        char *cid = 0;
        if(sd.ClientID[0]) cid = sd.ClientID;
        Init(sd.Interface,0,0,cid,sd.Append);
        SetConformity(sd.Flags);
        if(sd.HostName[0]) SetAuthName(sd.HostName);
        if(sd.ClassID[0]) SetClassID(sd.ClassID);
        SetPrefTime(sd.Time);
        SetOptions(sd.Options);
        AddRequestOptions(sd.RequestOptions);
        SetTimeouts(sd.TimeoutValue,sd.TimeoutValue);

        PrevIP[0] = 0;
        FILE *file;
        char fname[128];
        sprintf(fname,"%s/%s.cache",MainPath,TargetInterface);
        file = fopen(fname,"r");
        if(file) {
                fscanf(file,"%20s",PrevIP);
                if(inet_addr(PrevIP) == INADDR_NONE) PrevIP[0] = 0;
                fclose(file);
        };

        if(sd.IgnoreFlag) PrevIP[0] = 0;

        Retry = 0;
        State = sd.State;

        next = 0;
        prev = LastSession;
        if(LastSession == 0) FirstSession = this;
        else LastSession[0].next = this;
        LastSession = this;
}

DHCPClient::~DHCPClient()
{
        if(prev != 0) prev[0].next = next;
        else FirstSession = next;
        if(next != 0) next[0].prev = prev;
        else LastSession = prev;
}

void    DHCPClient::Last()
{
        if(prev != 0) prev[0].next = next;
        else FirstSession = next;
        if(next != 0) next[0].prev = prev;
        else LastSession = prev;

        next = 0;
        prev = LastSession;
        if(LastSession == 0) FirstSession = this;
        else LastSession[0].next = this;
        LastSession = this;
}

void    DHCPClient::First()
{
        if(prev != 0) prev[0].next = next;
        else FirstSession = next;
        if(next != 0) next[0].prev = prev;
        else LastSession = prev;

        prev = 0;
        next = FirstSession;
        if(FirstSession == 0) LastSession = this;
        else FirstSession[0].prev = this;
        FirstSession = this;
}




void    DHCPClient::Handler(int Op,char *s1,char *s2)
{
        int c;
        char *action = 0;
        char reason[64] = {""};

        switch(Op) {
                case DHCP_DISCOVER :
                        if(s1 == 0) action = "";
                        else action = ipname((unsigned char *) s1);
                        printf("%s: DHCPDISCOVER %s\n",TargetInterface,action);
                        break;
                case DHCP_OFFER :
                        printf("%s, %s: DHCPOFFER %s\n",TargetInterface,
                                                                   IPName(ipname((unsigned char *) s1)).Name,
                                                                   IPName(ipname((unsigned char *) s2)).Name);
                        break;
                case DHCP_DECLINE :
                        printf("%s: DHCPDECLINE %s\n",TargetInterface,
                                                                 IPName(ipname((unsigned char *) s1)).Name);
                        break;
                case DHCP_INFORM :
                        action = "INFORM";
                case DHCP_REQUEST_SELECT :
                        if(action == 0) action = "DHCPREQUEST, SELECT";
                case DHCP_REQUEST_RESELECT :
                        if(action == 0) action = "DHCPREQUEST, RESELECT";
                        printf("%s, %s: %s %s\n",TargetInterface,
                                                            IPName(ipname((unsigned char *) s1)).Name,
                                                            action,
                                                            IPName(ipname((unsigned char *) s2)).Name);
                        break;
                case DHCP_ACK :
                        printf("   IP         : %s\n",IPName(ipname(LocalIP)).Name);
                        printf("   SERVER     : %s\n",IPName(ipname(ServerIP)).Name);
                        printf("   HOST       : %s\n",HostName);
                        printf("   DOMAIN     : %s\n",DomainName);
                        printf("   RENEW      : %u\n",(unsigned int) RenewTime);
                        printf("   NETMASK    : %s\n",ipname(Netmask));
                        printf("   BROADCAST  : %s\n",ipname(Broadcast));

                        c = 0;
                        printf("   GATEWAY(s) : ");
                        while(c < 4 && memcmp(Routers[c],"\x0\x0\x0\x0",4) != 0) {
                                if(c > 0) printf(", %s",ipname(Routers[c]));
                                else printf("%s",ipname(Routers[c]));
                                c++;
                        };
                        puts("");

                        c = 0;
                        printf("   DNS        : ");
                        while(c < 4 && memcmp(DomainNameServers[c],"\x0\x0\x0\x0",4) != 0) {
                                if(c > 0) printf(", %s",ipname(DomainNameServers[c]));
                                else printf("%s",ipname(DomainNameServers[c]));
                                c++;
                        };
                        puts("");

                        strcpy(PrevIP,ipname(LocalIP));
                        WriteCache();
                        WriteInfo();
                        ExecScript("open");

                        if(WaitFlag) {
                                WaitFlag = 0;
                                WaitCount--;
                                if(WaitCount == 0) CFork();
                        };

                        break;
                case DHCP_NACK :
                        if(s1) {
                                memset(reason,0,64);
                                strncpy(reason,s1,63);
                        };
                        printf("%s: DHCPNACK: %s\n",TargetInterface,
                                                               reason);
//                        if(s2) Decline((unsigned char *) s2);
                        break;
                case DHCP_REQUEST_RENEW :
                        action = "RENEW";
                case DHCP_REQUEST_REBIND :
                        if(action == 0) action = "REBIND";
                        printf("%s, %s: DHCPREQUEST, %s %s\n",TargetInterface,
                                                                         IPName(ipname(ServerIP)).Name,
                                                                         action,
                                                                         IPName(ipname(LocalIP)).Name);
                        break;
                case DHCP_RENEW :
                        strcpy(PrevIP,ipname(LocalIP));
                        WriteCache();
                        WriteInfo();
                        ExecScript("update");

                        printf("   RENEW      : %u\n",(unsigned int) RenewTime);
                        break;
                case DHCP_RELEASE :
                        if(strcmp(LastAction,"open") == 0) {
                                ExecScript("close");
                        };
                        printf("%s, %s: DHCPRELEASE %s\n",TargetInterface,
                                                                     IPName(ipname(ServerIP)).Name,
                                                                     IPName(ipname(LocalIP)).Name);
                        break;
                case DHCP_INIT :
                        printf("%s, %s, %s, %s\n",TargetInterface,Interface,
                                                             HWName(FmtHW(HardwareAddress,HALength)).Name,
                                                             HWName(FmtHW(ClientID,CIDLength)).Name);
                        break;
                case DHCP_ERROR :
                        fprintf(stderr,"ERROR: %s\n",s1);
                        break;
                case DHCP_DEBUG :
                        fprintf(stderr,"DEBUG: %s\n",s1);
                        break;
        };
}

void            DHCPClient::Close()
{
        if(RenewTime != 0) {
                if(ReleaseFlag > 0) {
                        Release();
                };
                if(ReleaseFlag < 0) {
                        SetTimeouts(3000,1000);
                        int c = 0;
                        while(c < 3 && Renew() < 0) c++;
                };
        };

        if(strcmp(LastAction,"open") == 0) {
                ExecScript("close");
        };
}

int             DHCPClient::EventManager()
{
        DHCPClient *d;
        unsigned long wait,t,w;

        while(FirstSession != 0 && KillCount == 0) {
                d = FirstSession;
                t = GetSecs();
                wait = 0xFFFFFFFF;
                while(d) {
                        w = (d[0].*d[0].State) (t);
                        if(w == 0) break;
                        if(w < wait) wait = w;
                        d = d[0].next;
                };

#ifndef __WIN32__
                if(d) {
                        if(d[0].State != &DHCPClient::StateTransition) d[0].Last();
                        else d[0].First();
                }
                else {
                        AlarmCount = 0;
                        if(wait != 0xFFFFFFFF) {
/*                                itimerval tv;

                                tv.it_interval.tv_sec = 0;
                                tv.it_interval.tv_usec = 0;
                                tv.it_value.tv_sec = wait;
                                tv.it_value.tv_usec = 0;

                                setitimer(ITIMER_REAL,&tv,0);*/
//                                alarm(wait);
                                printf("Sleeping %d seconds...\n",(int) wait);
                        }
                        else {
                                puts("(Suspend)");
                        };

                        while(wait > 0 &&
                              AlarmCount == 0 &&
                              TriggerCount == 0 &&
                              KillCount == 0) {
                                wait = sleep(wait);
                        };
                };
#endif
                if(TriggerCount) {
                        d = FirstSession;
                        while(d) {
                                if(d[0].State == &DHCPClient::StateRenew ||
                                   d[0].State == &DHCPClient::StateRebind) {
                                        d[0].Timeout = 0;
                                };
                                d = d[0].next;
                        };
                        TriggerCount = 0;
                };
        };

        puts("Shutting down...");
        return 0;
}

unsigned long   DHCPClient::StateSleep(unsigned long t)
{
        if(AfterSleepState == 0) return 0xFFFFFFFF;

        unsigned long w = t - TickStart;
        if(w < StateTimeout) return StateTimeout - w;
        State = AfterSleepState;
        TickStart = t;
        return (this[0].*State) (t);
}

unsigned long   DHCPClient::StateDiscover(unsigned long)
{
        if(Retry >= 3) {
                if(WaitFlag) {
                        WaitFlag = 0;
                        WaitCount--;
                        if(WaitCount == 0) CFork();
                };
                
                Retry = 0;
                StateTimeout = 60;
                SleepState(&DHCPClient::StateReselect);
                return 0;
        };

        int ret;
        ret = Discover(0);
        if(ret > 0) {
                Retry = 0;
                MorphState(&DHCPClient::StateDiscoverSelect);
                return (this[0].*State) (0);
        };
        if(ret <= 0) Retry++;

        return 0;
}

unsigned long   DHCPClient::StateRediscover(unsigned long)
{
        if(Retry >= 3 || PrevIP[0] == 0) {
                Retry = 0;
                MorphState(&DHCPClient::StateDiscover);
                return 0;
        };

        int ret;
        ret = Discover(PrevIP);
        if(ret > 0) {
                Retry = 0;
                MorphState(&DHCPClient::StateRediscoverSelect);
                return (this[0].*State) (0);
        };
        if(ret <= 0) Retry++;

        return 0;
}

unsigned long   DHCPClient::StateInform(unsigned long)
{
        if(Retry >= 3) {
                Retry = 0;
                StateTimeout = 60;
                SleepState(&DHCPClient::StateInform);
                return 0;
        };
        
        int ret;
        ret = Inform();
        if(ret >= 0) SleepState(0);
        if(ret < 0) Retry++;

        return 0;
}

unsigned long   DHCPClient::StateDiscoverSelect(unsigned long)
{
        if(Retry >= 12) {
                Retry = 0;
                StateTimeout = 60;
                SleepState(&DHCPClient::StateReselect);
                return 0;
        };

        int ret;
        ret = Select(Retry / 3);
        if(ret > 0) {
                StateTimeout = (RenewTime * 6) / 8;
                Timeout = (RenewTime * 5) / 8;
                Retry = 0;
                ChangeState(&DHCPClient::StateRenew);
        };
        if(ret == 0) Retry += 3 - (Retry % 3);  //go to next
        if(ret < 0) Retry++;

        return 0;
}

unsigned long   DHCPClient::StateRediscoverSelect(unsigned long)
{
        if(Retry >= 3 || PrevIP[0] == 0) {
                Retry = 0;
                MorphState(&DHCPClient::StateDiscover);
                return 0;
        };

        int ret;
        ret = Select(0,PrevIP);
        if(ret > 0) {
                StateTimeout = (RenewTime * 6) / 8;
                Timeout = (RenewTime * 5) / 8;
                Retry = 0;
                ChangeState(&DHCPClient::StateRenew);
        };
        if(ret == 0) Retry = 3;
        if(ret < 0) Retry++;
        return 0;
}

unsigned long   DHCPClient::StateReselect(unsigned long)
{
        if(Retry >= 3 || PrevIP[0] == 0) {
                Retry = 0;
                MorphState(&DHCPClient::StateRediscover);
                return 0;
        };

        int ret;
        ret = Reselect(PrevIP);
        if(ret > 0) {
                StateTimeout = (RenewTime * 6) / 8;
                Timeout = (RenewTime * 5) / 8;
                Retry = 0;
                ChangeState(&DHCPClient::StateRenew);
        };
        if(ret == 0) {
#ifdef __RECORD__
                FILE *f = fopen("send.out","a+");
                int blen = sizeof(Packet);
                fwrite(&blen,4,1,f);
                fwrite(&Packet,blen,1,f);
                fclose(f);
#endif
                Retry = 3;
        };
        if(ret < 0) Retry++;
        return 0;
}

unsigned long   DHCPClient::StateTransition(unsigned long t)
{
        State = TransState;
        TickStart = t;
        return 0;
}

unsigned long   DHCPClient::StateRenew(unsigned long t)
{
        unsigned long w = t - TickStart;

        if(w >= StateTimeout) {
                Timeout = StateTimeout;
                StateTimeout = (RenewTime * 7) / 8;
                Retry = 0;
                MorphState(&DHCPClient::StateRebind);
                return 0;
        };

        if(w >= Timeout) {
                if(Retry >= 3 && (StateTimeout - w) > 60) {
                        Retry = 0;
                        unsigned long delta;
                        delta = (StateTimeout - w) / 2;
                        Timeout = w + delta;
                }
                else {
                        int ret;
                        ret = Renew();

                        if(ret == 0 && NackFlag != 0) {
                                ExecScript("close");
                                ExecScript("init");
                                Retry = 0;
                                MorphState(&DHCPClient::StateReselect);
                                return 0;
                        };

                        if(ret > 0) {
                                StateTimeout = (RenewTime * 6) / 8;
                                Timeout = (RenewTime * 5) / 8;
                                Retry = 0;
                                ChangeState(&DHCPClient::StateRenew);
                        };
                        if(ret <= 0) Retry++;

                        return 0;
                };
        }

        return Timeout - w;
}

unsigned long   DHCPClient::StateRebind(unsigned long t)
{
        unsigned long w = t - TickStart;

        if(w >= StateTimeout) {
                ExecScript("close");
                ExecScript("init");
                Retry = 0;
                MorphState(&DHCPClient::StateReselect);
                return 0;
        };

        if(w >= Timeout) {
                if(Retry >= 3 && (StateTimeout - w) > 60) {
                        Retry = 0;
                        unsigned long delta;
                        delta = (StateTimeout - w) / 2;
                        Timeout = w + delta;
                }
                else {
                        int ret;
                        ret = Rebind();

                        if(ret == 0 && NackFlag != 0) {
                                ExecScript("close");
                                ExecScript("init");
                                Retry = 0;
                                MorphState(&DHCPClient::StateReselect);
                                return 0;
                        };

                        if(ret > 0) {
                                StateTimeout = (RenewTime * 6) / 8;
                                Timeout = (RenewTime * 5) / 8;
                                Retry = 0;
                                ChangeState(&DHCPClient::StateRenew);
                        };
                        if(ret <= 0) Retry++;

                        return 0;
                };
        }

        return Timeout - w;
}

int     DHCPClient::cmd_alias(char *s)
{
        struct Alias {
                const char *  str;
                char    real;
        };

        static Alias cmd[] = {
                              {"replybroadcasted",'a'},
                              {"requestbroadcasted",'b'},
                              {"ipcheck",'c'},
                              {"ignoreipcache",'d'},
                              {"scriptext",'e'},
                              {"help",'h'},
                              {"inform",'i'},
                              {"replyunicasted",'n'},
                              {"setscriptvar",'p'},
                              {"leasetime",'r'},
                              {"test",'t'},
                              {"version",'v'},
                              {"wait",'w'},
                              {"appendclientid",'A'},
                              {"baseinterface",'B'},
                              {"clientid",'C'},
                              {"deleteipcache",'D'},
                              {"hostname",'h'},
                              {"classid",'I'},
                              {"dhcpoption",'O'},
                              {"norelease",'R'},
                              {"nocollectoffers",'W'},
                              {"kill",'k'},
                              {"renew",'l'},
                              {"nackresets",'N'},
                              {"requestoption",'o'},
                              {0}};

        Alias * ptr = cmd;
        while(ptr[0].str) {
                if(strcmp(ptr[0].str,s) == 0) {
                        s[0] = ptr[0].real;
                        s[1] = 0;
                        return 1;
                };
                ptr++;
        };

        return 0;
}

int     DHCPClient::ProcessCmdLine(char * (*iter)(),StartupData *proto)
{

#ifndef __WIN32__
        static char help_msg[] = "Usage: dhcpxd -k|--kill [interface] | -l|--renew [interface] | [-abcdehinprtwABCDHIOR] interface\n\
Parameters:\n\
        -a              Always request broadcasted replies.\n\
        -b              Always broadcast requests.\n\
        -c              Check to see if IP address already in use.\n\
        -d              Discover any IP address.\n\
                        This simply ignores the cached IP address for\n\
                        the initial requesting sequence.\n\
        -e extension    Use the specified extension for script execution.\n\
                        Default is \"def\".\n\
        -h              This help screen.\n\
        -i              Initiate inform session.\n\
                        Use this to get the information on statically.\n\
                        configured IPs.\n\
        -k iface        Kill a session.\n\
        -l iface        Trigger renew state on interface(s).\n\
        -n              Never request broadcasted replies.\n\
                        Suggested technique for cable modems.\n\
                        However, you might always want this anyways.\n\
                        Using this options is STRONGLY recommended.\n\
        -o number       Request the option represented by the `number` to\n\
                        be sent in responses. Might be needed for the WINS.\n\
                        option.\n\
        -p VAR=VALUE    Supply the specified parameter(s) to scripts.\n\
        -r number       Supply the specified renew time (hours) in requests.\n\
                        Default is 24.\n\
                        NOTE: This is almost always overridden by the server.\n\
        -t              Test DHCP mechanisms on specific interface.\n\
                        Useful for determining the operating mode of the\n\
                        DHCP server.\n\
        -v              Display version information.\n\
        -w              Wait for the session to succeed or timeout.\n\
        -A client_id    Append the specified client identifier to default.\n\
        -B iface        Use the specified base interface for requests.\n\
                        Default is the specified target interface.\n\
        -C client_id    Use the specified client identifier.\n\
                        Default is the hardware MAC address.\n\
        -D              Delete the cache before starting session.\n\
        -H hostname     Supply the specified host name in requests.\n\
        -I string       Supply the specified class identifier in requests.\n\
        -N              Upon receiving NACK force into reselect.\n\
        -O num string   Supply the specified option in requests.\n\
                        Usage: -O 89 JerryWalker\n\
        -R              Do not release the IP address.\n\
                        This is useful for keeping the same IP address.\n\
                        If this option is specified more than once, the\n\
                        client will renew the address before shutting down.\n\
        -T num          Set timeout value in seconds.\n\
        -W              Do not wait when collecting DHCPOFFER responses.\n\
                        Useful for making the protocol \"speedy\".\n\
";
#else
        static char *help_msg = "";
#endif

        static StartupData *first = 0;
        static StartupData *last = 0;

        if(iter == 0) {
                StartupData *d,*temp;

                d = first;
                while(d) {
                        new DHCPClient(*d);
                        temp = d;
                        d = d[0].next;
                        delete temp;
                };
                first = 0;
                last = 0;
                return 0;
        };

        char *          par = 0;
        unsigned char * ptr;
        int             op,c,n,fl;

        StartupData sd;

        if(proto) {
                memcpy(&sd,&proto[0],sizeof(sd));
        }
        else {
                sd.next = 0;

                sd.Interface[0] = 0;
                sd.Flags = 0;
                sd.ClientID[0] = 0;
                sd.Append = 0;
                sd.HostName[0] = 0;
                sd.ClassID[0] = 0;
                sd.Options[0] = 255;
                sd.RequestOptions[0] = 0;
                sd.Time = 24*3600;
                sd.TimeoutValue = 3000;

                sd.TargetInterface[0] = 0;
                strcpy(sd.ScriptExtension,"def");
                sd.State = 0;
                sd.Parameters[0] = 0;

                sd.TestFlag = 0;
                sd.WaitFlag = 0;
                sd.IgnoreFlag = 0;
                sd.ReleaseFlag = 1;
                sd.DeleteFlag = 0;
                sd.NackFlag = 0;
        };

        while(1) {
                if(par == 0) {
                        par = iter();
                        if(par == 0) return 255;
                        if(par[0] == 0) return 0;
                }
                else par = iter();
                if(par == 0 || par[0] == 0) break;

                if(strcmp(par,"}") == 0) {
                        if(proto != 0) return 255;
                        printf("%s: } unexpected here.",_argv[0]);
                        return 24;
                };

                if(strcmp(par,"{") == 0) {
                        while(1) {
                                c = ProcessCmdLine(iter,&sd);
                                if(c == 255) return 0;
                                if(c) return c;
                        };
                };

                if(par[0] != '-') {
                        strncpy(sd.TargetInterface,par,16);
                        continue;
                };

                char sw[128];
                char *swpar = &sw[1];
                strcpy(sw,par);
                while(swpar[0]) {
                        switch(swpar[0]) {
                                case 'a' :
                                        sd.Flags |= DHCP_C_REQ_BCAST_REPLY_ALWAYS;
                                        break;
                                case 'b' :
                                        sd.Flags |= DHCP_C_SEND_BCAST_PACKETS;
                                        break;
                                case 'c' :
                                        sd.Flags |= DHCP_C_CHECK_IP_USE;
                                        break;
                                case 'd' :
                                        sd.IgnoreFlag = 1;
                                        break;
                                case 'e' :
                                        par = iter();
                                        if(par == 0 || par[0] == 0) {
                                                printf("%s: Expected parameter missing (-e).\n",_argv[0]);
                                                return 10;
                                        };
                                        strncpy(sd.ScriptExtension,par,16);
                                        break;
                                case 'h' :
                                        puts(help_msg);

                                        while(1) {
                                                par = iter();
                                                if(par == 0) return 255;
                                                if(par[0] == 0) return 0;
                                        };
                                case 'i' :
                                        sd.State = &DHCPClient::StateInform;
                                        break;
                                case 'k' :
                                        par = iter();
                                        if(par != 0 && par[0] == 0) par = 0;
                                        return KillSession(par);
                                case 'l' :
                                        par = iter();
                                        if(par != 0 && par[0] == 0) par = 0;
                                        return RenewSession(par);
                                case 'n' :
                                        sd.Flags |= DHCP_C_REQ_BCAST_REPLY_NEVER;
                                        break;
                                case 'o' :
                                        par = iter();
                                        if(par == 0 || par[0] == 0) {
                                                printf("%s: Expected parameter missing (-o).\n",_argv[0]);
                                                return 10;
                                        };
                                        c = atoi(par);
                                        if(c <= 0 || c >= 255) {
                                                printf("%s: Invalid non-zero numerical expression - %s\n",_argv[0],par);
                                                return 10;
                                        };
                                        ptr = sd.RequestOptions;
                                        while(*ptr && c != *ptr) ptr++;
                                        if(!*ptr) {
                                                *ptr = (unsigned char) c;
                                                ptr++;
                                                *ptr = 0;
                                        };
                                        break;
                                case 'p' :
                                        par = iter();
                                        if(par == 0 || par[0] == 0) {
                                                printf("%s: Expected parameter missing (-p).\n",_argv[0]);
                                                return 10;
                                        };
                                        if((strlen(par) + 1) < (128 - strlen(sd.Parameters))) {
                                                strcat(sd.Parameters,par);
                                                strcat(sd.Parameters,"\n");
                                        };
                                        break;
                                case 'r' :
                                        par = iter();
                                        if(par == 0 || par[0] == 0) {
                                                printf("%s: Expected parameter missing (-r).\n",_argv[0]);
                                                return 10;
                                        };
                                        sd.Time = atoi(par);
                                        if(sd.Time == 0) {
                                                printf("%s: Invalid non-zero numerical expression - %s\n",_argv[0],par);
                                                return 10;
                                        };
                                        sd.Time *= 3600UL;
                                        break;
                                case 't' :
                                        sd.TestFlag = 1;
                                        break;
                                case 'v' :
                                        printf("dhcpxd (%s)\n",DHCPXD_VERSION);
                                        while(1) {
                                                par = iter();
                                                if(par == 0) return 255;
                                                if(par[0] == 0) return 0;
                                        };
                                case 'w' :
                                        sd.WaitFlag = 1;
                                        WaitCount++;
                                        break;
                                case 'B' :
                                        par = iter();
                                        if(par == 0 || par[0] == 0) {
                                                printf("%s: Expected parameter missing (-B).\n",_argv[0]);
                                                return 10;
                                        };

                                        strncpy(sd.Interface,par,16);
                                        break;
                                case 'A' :
                                        sd.Append = 1;
                                case 'C' :
                                        if(swpar[0] == 'C') sd.Append = 0;

                                        par = iter();
                                        if(par == 0 || par[0] == 0) {
                                                printf("%s: Expected parameter missing (-A or -C).\n",_argv[0]);
                                                return 10;
                                        };
                                        strncpy(sd.ClientID,par,32);
                                        break;
                                case 'D' :
                                        sd.DeleteFlag = 1;
                                        break;
                                case 'H' :
                                        par = iter();
                                        if(par == 0 || par[0] == 0) {
                                                printf("%s: Expected parameter missing (-H).\n",_argv[0]);
                                                return 10;
                                        };

                                        strncpy(sd.HostName,par,64);
                                        break;
                                case 'I' :
                                        par = iter();
                                        if(par == 0 || par[0] == 0) {
                                                printf("%s: Expected parameter missing (-I).\n",_argv[0]);
                                                return 10;
                                        };

                                        strncpy(sd.ClassID,par,64);
                                        break;
                                case 'N' :
                                        sd.NackFlag = 1;
                                        break;
                                case 'O' :
                                        par = iter();
                                        if(par == 0 || par[0] == 0) {
                                                printf("%s: Expected parameter missing (-O).\n",_argv[0]);
                                                return 10;
                                        };

                                        op = atoi(par);
                                        if(op == 0) {
                                                printf("%s: Invalid numerical option - %s.\n",_argv[0],par);
                                                return 10;
                                        };

                                        par = iter();
                                        if(par == 0 || par[0] == 0) {
                                                printf("%s: Expected parameter missing (-O).\n",_argv[0]);
                                                return 10;
                                        };

                                        c = 0;
                                        while(sd.Options[c] != 255) {
                                                c++;
                                                c += sd.Options[c];
                                                c++;
                                        };

                                        if((int) (128 - c) >= (int) (strlen(par) + 3)) {
                                                sd.Options[c] = op;
                                                c++;
                                                sd.Options[c] = strlen(par);
                                                c++;
                                                memcpy(&sd.Options[c],par,sd.Options[c - 1]);
                                                c += sd.Options[c - 1];
                                                sd.Options[c] = 255;
                                        };

                                        break;
                                case 'R' :
                                        sd.ReleaseFlag--;
                                        break;
                                case 'T' :
                                        par = iter();
                                        if(par == 0 || par[0] == 0) {
                                                printf("%s: Expected parameter missing (-T).\n",_argv[0]);
                                                return 100;
                                        };

                                        sd.TimeoutValue = atoi(par) * 1000;
                                        break;
                                case 'W' :
                                        sd.Flags |= DHCP_C_NOCOLLECT_OFFERS;
                                        break;
                                case '-' :
//                                        if(strcmp(&swpar[1],"list") == 0) {
//                                        };

                                        if(cmd_alias(&swpar[1]) == 0) {
                                                printf("%s: invalid option - %s.\n",_argv[0],&swpar[1]);
                                                return 2;
                                        };
                                        break;
                                default:
                                        printf("%s: unknown option - %c.\n",_argv[0],swpar[0]);
                                        puts("Use -h for help.");
                                        return 2;
                        };

                        swpar++;
                };
        };

        if(sd.TargetInterface[0] == 0) {
                printf("%s: No interface(s) specified.\n",_argv[0]);
                return 2;
        };

        par = sd.TargetInterface;
        while(par[0] != 0 && par[0] != ',') par++;
        n = 1;
        fl = 0;
        if(par[0]) {
                par[0] = 0;
                par++;
                n = atoi(par);
                if(n == 0) {
                        printf("%s: Bad number.\n",_argv[0]);
                        return 24;
                };
                sd.Append = 1;
                fl = 1;
        };

        if(_argc > 1) sprintf(PIDFile,"/var/run/dhcpxd.%s.pid",sd.TargetInterface);

        if(sd.Interface[0] == 0) {
                strcpy(sd.Interface,sd.TargetInterface);
        };

        if(sd.State == 0) {
                sd.State = &DHCPClient::StateReselect;
        };

        c = 0;
        while(c < n) {
                if(fl) {
                        sprintf(par - 1,":%d",c);
                        sprintf(sd.ClientID,"%02d",c);
                };

                if(sd.DeleteFlag) {
                        char fname[128];
                        sprintf(fname,"%s/%s.cache",MainPath,sd.TargetInterface);
                        remove(fname);
                };

                if(sd.TestFlag) {
                        DHCPClient *d = new DHCPClient(sd);

                        int ret = d[0].Test();
                        delete d;
                        return ret;
                };

                StartupData *d = first;
                while(d) {
                        if(strcmp(d[0].TargetInterface,sd.TargetInterface) == 0) {
                                printf("%s: More than one session configures the same interface.\n",_argv[0]);
                        };

                        if(sd.Append && d[0].Append) {
                                if(strcmp(d[0].Interface,sd.Interface) == 0 &&
                                   strcmp(d[0].ClientID,sd.ClientID) == 0) {
                                        printf("%s: More than one session uses the same client ID.\n",_argv[0]);
                                };
                        };
                        if(sd.Append == 0 && d[0].Append == 0) {
                                if(strcmp(d[0].ClientID,sd.ClientID) == 0) {
                                        printf("%s: More than one session uses the same client ID.\n",_argv[0]);
                                };
                        };

                        d = d[0].next;
                };

                if(last) {
                        last[0].next = new StartupData;
                        last = last[0].next;
                }
                else {
                        last = new StartupData;
                        first = last;
                };
                memcpy((void *) last,(void *) &sd,sizeof(sd));

                sd.HostName[0] = 0;
                c++;
        };

        return 0;
}

int     DHCPClient::KillSession(char *which)
{
        FILE *f;
        int pid;
        int lfd = LockX();

        if(which) sprintf(PIDFile,"/var/run/dhcpxd.%s.pid",which);

        f = fopen(PIDFile,"r");
        if(f) {
                pid = 0;
                fscanf(f,"%d",&pid);
                fclose(f);

                if(pid != 0 && (f = fopen(PIDFile,"a+")) != 0) {
                        fprintf(f," %d",getpid());
                        fclose(f);
                        UnlockX(lfd);

                        printf("Terminating dhcpxd (PID %d)...\n",pid);
                        KillCount = 0;
                        kill(pid,SIGTERM);

                        while(KillCount == 0) pause();

                        return 68;
                };

                UnlockX(lfd);
                return 67;
        };

        UnlockX(lfd);
        return 67;
}

int     DHCPClient::RenewSession(char *which)
{
        FILE *f;
        int pid;
        int lfd = LockX();

        if(which) sprintf(PIDFile,"/var/run/dhcpxd.%s.pid",which);

        f = fopen(PIDFile,"r");
        if(f) {
                pid = 0;
                fscanf(f,"%d",&pid);
                fclose(f);

                if(pid) {
                        UnlockX(lfd);

                        kill(pid,SIGUSR1);
                        return 68;
                };

                UnlockX(lfd);
                return 67;
        };

        UnlockX(lfd);
        return 67;
}

int     DHCPClient::Main(char * (*iter)())
{
#ifndef __WIN32__
        Signal(SIGUSR1,TriggerHandler);
        Signal(SIGALRM,AlarmHandler);
        Signal(SIGTERM,KillHandler);
#endif

        FILE *f;
        int ret,pid,lfd;

        while((ret = ProcessCmdLine(iter,0)) == 0) {};
        if(ret != 255) return ret;

        lfd = LockX();
        f = fopen(PIDFile,"r");
        if(f) {
                pid = 0;
                fscanf(f,"%d",&pid);
                fclose(f);
                printf("dhcpxd: dhcpxd is already running as pid %d\n",pid);
                UnlockX(lfd);
                return 67;
        };
        UnlockX(lfd);

        ProcessCmdLine(0,0);

        if(FirstSession == 0) return 0;

        if(WaitCount == 0) CFork();

        ret = EventManager();

        while(FirstSession) {
                FirstSession[0].Close();
                delete FirstSession;
        };

        lfd = LockX();
        f = fopen(PIDFile,"r");
        if(f) {
                int c;
                c = fscanf(f,"%d",&pid);
                while(c != EOF && c > 0) {
                        kill(pid,SIGTERM);
                        c = fscanf(f,"%d",&pid);
                };
                fclose(f);
        };
        remove(PIDFile);
        UnlockX(lfd);

        return ret;
}




#ifdef __WIN32__
main()
{
        WSADATA wsad;
        WSAStartup(0x0101,&wsad);

        DHCPClient d;

        d.SetPrimeIP("24.64.221.215");
        d.SetAuthName("Experimental");

        d.SetPrefTime(24*3600);
        d.SetConformity(DHCP_C_REQ_BCAST_REPLY_NEVER);
        d.SetTimeouts(60000,3000);
        d.Test();

        return 0;
}

#else

FILE *  _cmdstream;

char *  streamIterator()
{
        if(_cmdstream == 0) return 0;
        static char str[128] = {""};

        int     quotefl = 0;
        int     remfl = 0;
        int     initfl = 1;
        char *  ptr = str;
        int     c;
        while((c = fgetc(_cmdstream)) != EOF) {
                switch(c) {
                        case '\n' :
                                remfl = 0;
                                if(initfl != 0 && str[0] != 0) {
                                        str[0] = 0;
                                        return "";
                                };
                                if(initfl) break;
                                ungetc('\n',_cmdstream);
                                return str;
                        case '\t' :
                                c = ' ';
                        case ' ' :
                                if(initfl) break;
                                if(remfl) break;
                                if(quotefl) {
                                        if((ptr - str) >= 127) break;
                                        ptr[0] = c;
                                        ptr++;
                                        ptr[0] = 0;
                                        break;
                                };
                                return str;
                        case '#' :
                                if(quotefl) {
                                        if((ptr - str) >= 127) break;
                                        ptr[0] = c;
                                        ptr++;
                                        ptr[0] = 0;
                                        break;
                                };
                                remfl = 1;
                                break;
                        case '"' :
                                if(remfl) break;
                                quotefl ^= 1;
                                initfl = 0;
                                break;
                        case '\\' :
                                c = fgetc(_cmdstream);
                                if(c == EOF) break;
                        default:
                                if(remfl) break;

                                initfl = 0;
                                if((ptr - str) >= 127) break;
                                ptr[0] = c;
                                ptr++;
                                ptr[0] = 0;
                                break;
                };
        };

        fclose(_cmdstream);
        _cmdstream = 0;
        if(ptr != str) return str;
        return 0;
}


char *  argIterator()
{
        static int pos = 1;

        if(pos == _argc) return 0;
        return _argv[pos++];
}

main(int argc,char *argv[])
{
        _argv = argv;
        _argc = argc;

        if(argc == 1) {
                static char fname[128];
                sprintf(fname,"%s/dhcpxd.conf",MainPath);
                _cmdstream = fopen(fname,"r");
                if(_cmdstream == 0) {
                        printf("%s: %s: %s\n",_argv[0],fname,strerror(errno));
                        return 1;
                };
                printf("Processing %s\n",fname);
                int ret = DHCPClient::Main(streamIterator);
                if(_cmdstream) fclose(_cmdstream);
                return ret;
        }
        else {
                return DHCPClient::Main(argIterator);
        };                
}

#endif
